import pytest

from weatherflow4py.models.ws.websocket_request import (
    GeoStrikeListenStartMessage,
    LightingStrikeType,
)
from weatherflow4py.models.ws.websocket_response import (
    WebsocketResponseBuilder,
    AcknowledgementWS,
    ConnectionOpenWS,
    ObservationTempestWS,
    RapidWindWS,
    EventDataRapidWind,
    LightningStrikeEventWS,
)


def test_websocket_status(websocket_satus_invalid):
    resp = WebsocketResponseBuilder.build_response(websocket_satus_invalid)
    assert resp is None


def test_ws_strike(websocket_strike):
    lse: LightningStrikeEventWS = LightningStrikeEventWS.from_dict(websocket_strike)
    assert lse.unknown_fields == {
        "hub_sn": "HB-00061234",
        "serial_number": "ST-00081234",
        "source": "enhanced",
    }

    assert lse.energy == -182
    assert lse.epoch == 1710383159
    assert lse.distance_km == 13


def test_websocket_messages(websocket_messages):
    msg0 = WebsocketResponseBuilder.build_response(websocket_messages[0])
    assert isinstance(msg0, ConnectionOpenWS)
    msg1 = WebsocketResponseBuilder.build_response(websocket_messages[1])
    assert isinstance(msg1, AcknowledgementWS)

    with pytest.raises(Exception) as exc_info:
        WebsocketResponseBuilder.build_response(websocket_messages[2])
        assert exc_info is ValueError
        assert exc_info.type == "Taco"
        assert "An error occurred" in str(exc_info.value)
        print(exc_info)

    msg3 = WebsocketResponseBuilder.build_response(websocket_messages[3])
    assert isinstance(msg3, ObservationTempestWS)
    msg4 = WebsocketResponseBuilder.build_response(websocket_messages[4])
    assert isinstance(msg4, ObservationTempestWS)
    msg5 = WebsocketResponseBuilder.build_response(websocket_messages[5])
    assert isinstance(msg5, ObservationTempestWS)

    assert msg3.device_id == 211522
    assert msg3.epoch == 1709130791


def test_winds(websocket_winds):
    for msg in websocket_winds:
        wind = WebsocketResponseBuilder.build_response(msg)
        assert isinstance(wind, RapidWindWS)
        assert isinstance(wind.ob, EventDataRapidWind)


@pytest.mark.parametrize(
    "strike_type, expected",
    [
        (LightingStrikeType.ALL, "all"),
        (LightingStrikeType.CLOUD_TO_GROUND, "cg"),
        (LightingStrikeType.CLOUD_TO_CLOUD, "ic"),
        (None, None),
    ],
)
def test_geo_strike_listen_start_message_to_dict_with_different_strike_types(
    strike_type, expected
):
    message = GeoStrikeListenStartMessage(
        lat_min=37.28,
        lat_max=41.32,
        lon_min=-101.76,
        lon_max=-91.00,
        strike_type=strike_type,
    )
    message_dict = message.to_dict()
    assert message_dict["type"] == "geo_strike_listen_start"
    assert message_dict["lat_min"] == 37.28
    assert message_dict["lat_max"] == 41.32
    assert message_dict["lon_min"] == -101.76
    assert message_dict["lon_max"] == -91.00
    assert "id" in message_dict
    if expected is not None:
        assert message_dict["strike_type"] == expected
    else:
        assert "strike_type" not in message_dict



import asyncio
import os

from weatherflow4py.api import WeatherFlowRestAPI


from dotenv import load_dotenv

import logging

logging.basicConfig(level=logging.DEBUG)


async def main():
    load_dotenv()  # load environment variables from .env file

    token = os.getenv("TOKEN")
    async with WeatherFlowRestAPI(token) as api:
        data = await api.get_all_data()
        print(data)

    while True:
        async with WeatherFlowRestAPI(token) as api:
            station_response = await api.async_get_stations()
            for station in station_response.stations:
                print(station.name)
                print(station.station_id)
                print(station.public_name)
                print(station.latitude)
                print(station.longitude)
                print(station.elevation)
                print(station.devices)
                print("\n")
                forecast = await api.async_get_forecast(station_id=station.station_id)
                print(forecast.units)

                obs = await api.async_get_observation(station_id=station.station_id)

                print(obs.station_id)
                print(obs.obs[0].air_temperature)

                # obs = await api.async_get_device_observations()
                # device_obs = await api.async_get_device_observations(
                #     station.outdoor_devices[0].device_ids
                # )
                # print(device_obs.precipitation_type)
            await asyncio.sleep(5)


if __name__ == "__main__":
    asyncio.run(main())



import json
import os

import pytest
from aioresponses import aioresponses
from weatherflow4py.api import WeatherFlowRestAPI

dir_path = os.path.dirname(os.path.realpath(__file__))


def load_fixture(file_name):
    with open(os.path.join(dir_path, file_name)) as json_file:
        return json.load(json_file)


@pytest.fixture
def websocket_messages():
    return load_fixture("fixtures/ws/websocket_messages.json")


@pytest.fixture
def websocket_winds():
    return load_fixture("fixtures/ws/ws_winds.json")


@pytest.fixture
def websocket_strike():
    return load_fixture("fixtures/ws/ws_strike.json")


@pytest.fixture
def websocket_satus_invalid():
    return load_fixture("fixtures/ws/ws_status.json")


@pytest.fixture
def ws_obs_air():
    return load_fixture("fixtures/ws_obs_air.json")


@pytest.fixture
def ws_obs_sky():
    return load_fixture("fixtures/ws_obs_sky.json")


@pytest.fixture
def ws_obs_st():
    return load_fixture("fixtures/ws_obs_st.json")


@pytest.fixture
def ws_obs_st_3x():
    return load_fixture("fixtures/ws_obs_st_3x.json")


@pytest.fixture
def obs_st_json():
    return load_fixture("fixtures/obs_st.json")


@pytest.fixture
def rest_betterforecast_1():
    return load_fixture("fixtures/rest/betterforecast/forecast.json")


@pytest.fixture
def rest_betterforecast_2():
    return load_fixture("fixtures/rest/betterforecast/forecast2.json")


@pytest.fixture
def rest_betterforecast_3():
    return load_fixture("fixtures/rest/betterforecast/forecast3.json")


@pytest.fixture
def rest_betterforecast_4():
    return load_fixture("fixtures/rest/betterforecast/forecast4.json")


@pytest.fixture
def rest_betterforecast_5():
    return load_fixture("fixtures/rest/betterforecast/forecast5.json")


@pytest.fixture
def rest_betterforecast_6():
    return load_fixture("fixtures/rest/betterforecast/forecast6.json")


@pytest.fixture
def rest_betterforecast_missing_fields():
    return load_fixture("fixtures/rest/betterforecast/forecast_missing_fields.json")


@pytest.fixture
def rest_station_observation1():
    return load_fixture(
        "fixtures/rest/observations/station_id/station_observation1.json"
    )


@pytest.fixture
def rest_station_observation2():
    return load_fixture(
        "fixtures/rest/observations/station_id/station_observation2.json"
    )


@pytest.fixture
def rest_station_observation_no_precip():
    return load_fixture(
        "fixtures/rest/observations/station_id/station_observation_no_precip.json"
    )


@pytest.fixture
def rest_station_observation_no_wind():
    return load_fixture(
        "fixtures/rest/observations/station_id/station_observation_no_wind.json"
    )


@pytest.fixture
def rest_device_observation_1():
    return load_fixture("fixtures/rest/observations/device_id/1.json")


@pytest.fixture
def rest_device_observation_2():
    return load_fixture("fixtures/rest/observations/device_id/2.json")


@pytest.fixture
def unauthorized_json():
    return load_fixture("fixtures/rest/401.json")


@pytest.fixture
def rest_internal_error_json():
    return load_fixture("fixtures/rest/internal_error.json")


@pytest.fixture
def rest_station_json():
    return load_fixture("fixtures/rest/stations/station_id/station_id.json")


@pytest.fixture
def rest_stations_json():
    return load_fixture("fixtures/rest/stations/stations.json")


@pytest.fixture
def rest_stations_with_errors_json():
    return load_fixture("fixtures/rest/stations/stations_with_errors.json")


@pytest.fixture
async def mock_aioresponse():
    with aioresponses() as m:
        yield m


@pytest.fixture
async def weather_api():
    api = WeatherFlowRestAPI("your_api_token")
    async with api as api_instance:
        yield api_instance



import asyncio
import os
from collections import Counter
from dotenv import load_dotenv

from weatherflow4py.models.ws.websocket_request import (
    ListenStartMessage,
)
from weatherflow4py.models.ws.websocket_response import (
    RapidWindWS,
    ObservationTempestWS,
)
from weatherflow4py.ws import WeatherFlowWebsocketAPI
import logging
from pprint import pprint

NUMBER_OF_MESSAGES = 5
REQUIRED_MESSAGE_TYPES = {"obs_st", "rapid_wind"}

ws_logger = logging.getLogger("websockets.client")
ws_logger.setLevel(logging.INFO)

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

message_counter = Counter()
received_message_types = set()


def invalid_data_cb(data):
    print("Invalid data ❗️ received:", data)


def wind_cb(data: RapidWindWS):
    print("Wind 🍃 data received:", data)
    message_counter["rapid_wind"] += 1
    received_message_types.add("rapid_wind")
    print_message_count()


def lightning_cb(data):
    print("Lightning ⚡️ data received:", data)
    message_counter["evt_strike"] += 1
    received_message_types.add("evt_strike")
    print_message_count()


def obs_cb(data: ObservationTempestWS):
    print("Observation 🔎️ received")
    message_counter["obs_st"] += 1
    received_message_types.add("obs_st")
    print_message_count()


def precipitation_cb(data):
    print("Precipitation 🌧️ data received:", data)
    message_counter["evt_precip"] += 1
    received_message_types.add("evt_precip")
    print_message_count()


def print_message_count():
    total_messages = sum(message_counter.values())
    remaining = max(0, NUMBER_OF_MESSAGES - total_messages)
    print(
        f"Received {total_messages}/{NUMBER_OF_MESSAGES} messages. {remaining} remaining."
    )
    print(f"Message types received: {received_message_types}")


async def main():
    load_dotenv()  # load environment variables from .env file

    token = os.getenv("TOKEN")
    device = os.getenv("DEVICE")

    api = WeatherFlowWebsocketAPI(token, device)
    api.register_observation_callback(obs_cb)
    api.register_wind_callback(wind_cb)
    api.register_lightning_callback(lightning_cb)
    api.register_precipitation_callback(precipitation_cb)

    await api.connect()
    # Start most messages
    await api.send_message_and_wait(ListenStartMessage(device_id=device))
    # Start wind messages
    # await api.send_message_and_wait(RapidWindListenStartMessage(device_id=device))

    while sum(
        message_counter.values()
    ) < NUMBER_OF_MESSAGES or not REQUIRED_MESSAGE_TYPES.issubset(
        received_message_types
    ):
        await asyncio.sleep(1)

    print("-" * 80)
    await api.stop_all_listeners()
    print("-" * 80)
    # await api.send_message_and_wait(ListenStopMessage(device_id=device))
    # await api.send_message_and_wait(RapidWindListenStopMessage(device_id=device))

    print("Received all required message types and reached message limit.")
    print("Final message count:")
    pprint(dict(message_counter))
    print("DATA::")
    pprint(api.messages)

    # await api.close()
    print("Connection closed.")


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(main())
    except KeyboardInterrupt:
        logger.info("Received keyboard interrupt. Shutting down...")
    finally:
        pending = asyncio.all_tasks(loop=loop)
        try:
            loop.run_until_complete(
                asyncio.wait_for(
                    asyncio.gather(*pending, return_exceptions=True), timeout=10
                )
            )
        except TimeoutError:
            logger.warning("Listen task cancellation timed out or was cancelled")

        # Keep the loop running until Ctrl+C is hit
        try:
            while True:
                loop.run_forever()
        except KeyboardInterrupt:
            logger.info("Received keyboard interrupt. Exiting...")
        finally:
            loop.close()



import pytest
from aioresponses import aioresponses  # Make sure to import aioresponses

from weatherflow4py.api import WeatherFlowRestAPI
from weatherflow4py.exceptions import TokenError
from weatherflow4py.models.rest.device import DeviceObservationREST
from weatherflow4py.models.rest.forecast import WeatherDataForecastREST
from aiohttp import ClientResponseError

from weatherflow4py.models.rest.stations import StationsResponseREST
from weatherflow4py.models.rest.unified import WeatherFlowDataREST


@pytest.mark.asyncio
async def test_api_calls(
    rest_betterforecast_1,
    rest_station_json,
    rest_stations_json,
    rest_station_observation2,
    rest_device_observation_1,
):
    import logging

    logging.basicConfig(level=logging.DEBUG)

    with aioresponses() as mock:
        mock.get(
            "https://swd.weatherflow.com/swd/rest/stations?token=mock_token",
            payload=rest_stations_json,
        )
        mock.get(
            "https://swd.weatherflow.com/swd/rest/better_forecast?station_id=24432&token=mock_token",
            payload=rest_betterforecast_1,
        )
        mock.get(
            "https://swd.weatherflow.com/swd/rest/observations/station?station_id=24432&token=mock_token",
            payload=rest_station_json,
        )

        mock.get(
            "https://swd.weatherflow.com/swd/rest/observations/device/123456?token=mock_token",
            payload=rest_device_observation_1,
        )
        mock.get(
            "https://swd.weatherflow.com/swd/rest/observations/station/24432?token=mock_token",
            payload=rest_station_observation2,
        )
        mock.get(
            "https://swd.weatherflow.com/swd/rest/observations/station/24432?token=mock_token",
            payload=rest_station_observation2,
        )

        sor = StationsResponseREST.from_dict(rest_stations_json)
        assert isinstance(sor, StationsResponseREST)
        sor2 = StationsResponseREST.from_dict(rest_station_json)
        assert isinstance(sor2, StationsResponseREST)

        dob = DeviceObservationREST.from_dict(rest_device_observation_1)
        assert isinstance(dob, DeviceObservationREST)
        wf = WeatherDataForecastREST.from_dict(rest_betterforecast_1)
        assert isinstance(wf, WeatherDataForecastREST)

        async with WeatherFlowRestAPI("mock_token") as api:
            data: WeatherFlowDataREST = await api.get_all_data(
                get_device_observations=True
            )

            assert data[24432].station.station_id == 24432


@pytest.mark.asyncio
async def test_api_calls_unauthorized(unauthorized_json):
    with aioresponses() as mock:
        mock.get(
            "https://swd.weatherflow.com/swd/rest/stations?token=mock_token",
            payload=unauthorized_json,
            status=401,
        )

        async with WeatherFlowRestAPI("mock_token") as api:
            # Expect an HTTPUnauthorized exception
            with pytest.raises(ClientResponseError):
                await api.async_get_stations()


@pytest.mark.asyncio
async def test_bad_token():
    with pytest.raises(TokenError):
        WeatherFlowRestAPI(None)  # type: ignore[invalid-argument-type]



import asyncio
import os

import websockets
import json
from weatherflow4py.models.ws.websocket_response import WebsocketResponseBuilder
from dotenv import load_dotenv


async def listen_to_tempest(device_id, access_token):
    uri = f"wss://ws.weatherflow.com/swd/data?token={access_token}"

    async with websockets.connect(uri) as websocket:
        # Start listening
        listen_start = {
            "type": "listen_start",
            "device_id": device_id,
            "id": "unique_id_here",  # Generate a unique ID for your request
        }
        await websocket.send(json.dumps(listen_start))

        # Handle incoming messages
        async for message in websocket:
            data = json.loads(message)
            #
            # Use the ResponseBuilder to create an instance of the appropriate response class
            try:
                response = WebsocketResponseBuilder.build_response(data)
                print(response)
            except ValueError:
                print(f"INVALID:\n {message}")

                continue

        # Stop listening (optional, depending on your application logic)
        listen_stop = {
            "type": "listen_stop",
            "device_id": device_id,
            "id": "unique_id_here",  # The same or another unique ID
        }
        await websocket.send(json.dumps(listen_stop))


def main():
    load_dotenv()  # load environment variables from .env file

    token = os.getenv("TOKEN")
    device = os.getenv("DEVICE")
    # Replace 'your_device_id' and 'your_access_token' with your actual Device ID and Access Token
    asyncio.run(listen_to_tempest(device, token))


if __name__ == "__main__":
    main()



from weatherflow4py.models.ws.websocket_request import (
    ListenStartMessage,
    ListenStopMessage,
    GeoStrikeListenStartMessage,
    RapidWindListenStartMessage,
    RapidWindListenStopMessage,
    LightingStrikeType,
)


def test_listen_start_message_to_dict():
    message = ListenStartMessage(device_id="1110")
    message_dict = message.to_dict()
    assert message_dict["type"] == "listen_start"
    assert message_dict["device_id"] == "1110"
    assert "id" in message_dict


def test_listen_stop_message_to_dict():
    message = ListenStopMessage(device_id="1110")
    message_dict = message.to_dict()
    assert message_dict["type"] == "listen_stop"
    assert message_dict["device_id"] == "1110"
    assert "id" in message_dict


def test_geo_strike_listen_start_message_to_dict_without_strike_type():
    message = GeoStrikeListenStartMessage(
        lat_min=37.28, lat_max=41.32, lon_min=-101.76, lon_max=-91.00
    )
    message_dict = message.to_dict()
    assert message_dict["type"] == "geo_strike_listen_start"
    assert message_dict["lat_min"] == 37.28
    assert message_dict["lat_max"] == 41.32
    assert message_dict["lon_min"] == -101.76
    assert message_dict["lon_max"] == -91.00
    assert "id" in message_dict
    assert "strike_type" not in message_dict


def test_geo_strike_listen_start_message_to_dict_with_strike_type():
    message = GeoStrikeListenStartMessage(
        lat_min=37.28,
        lat_max=41.32,
        lon_min=-101.76,
        lon_max=-91.00,
        strike_type=LightingStrikeType.ALL,
    )
    message_dict = message.to_dict()
    assert message_dict["type"] == "geo_strike_listen_start"
    assert message_dict["lat_min"] == 37.28
    assert message_dict["lat_max"] == 41.32
    assert message_dict["lon_min"] == -101.76
    assert message_dict["lon_max"] == -91.00
    assert "id" in message_dict
    assert message_dict["strike_type"] == "all"


def test_listen_rapid_start_message_to_dict():
    message = RapidWindListenStartMessage(device_id="1110")
    message_dict = message.to_dict()
    assert message_dict["type"] == "listen_rapid_start"
    assert message_dict["device_id"] == "1110"
    assert "id" in message_dict


def test_listen_rapid_stop_message_to_dict():
    message = RapidWindListenStopMessage(device_id="1110")
    message_dict = message.to_dict()
    assert message_dict["type"] == "listen_rapid_stop"
    assert message_dict["device_id"] == "1110"
    assert "id" in message_dict



import json

import pytest

from weatherflow4py.models.rest.stations import StationsResponseREST
from weatherflow4py.models.rest.status import Status
from weatherflow4py.models.ws.custom_types import PrecipitationType
from weatherflow4py.models.rest.device import (
    DeviceObservationTempestREST,
    PrecipitationAnalysisType,
)
from weatherflow4py.models.rest.forecast import (
    WeatherDataForecastREST,
    Forecast,
    CurrentConditions,
    ForecastUnits,
)
from weatherflow4py.models.ws.obs import ObservationType
from weatherflow4py.models.rest.observation import ObservationStationREST, WetBulbFlag


def test_convert_json_to_observation(rest_station_observation1):
    try:
        obs_data = ObservationStationREST.from_dict(rest_station_observation1)
    except Exception as e:
        pytest.fail(f"Failed to convert JSON data to Observation: {e}")

    assert obs_data.unknown_fields == {}

    assert isinstance(obs_data, ObservationStationREST)
    assert obs_data.elevation == rest_station_observation1["elevation"]
    assert obs_data.is_public == rest_station_observation1["is_public"]
    assert obs_data.latitude == rest_station_observation1["latitude"]
    assert obs_data.longitude == rest_station_observation1["longitude"]
    assert obs_data.public_name == rest_station_observation1["public_name"]
    assert obs_data.station_id == rest_station_observation1["station_id"]
    assert obs_data.station_name == rest_station_observation1["station_name"]
    assert obs_data.timezone == rest_station_observation1["timezone"]

    for obs, obs_json in zip(obs_data.obs, rest_station_observation1["obs"]):
        assert obs.air_density == obs_json["air_density"]
        assert obs.air_temperature == obs_json["air_temperature"]
        assert obs.barometric_pressure == obs_json["barometric_pressure"]
        assert obs.brightness == obs_json["brightness"]
        assert obs.delta_t == obs_json["delta_t"]
        assert obs.dew_point == obs_json["dew_point"]
        assert obs.feels_like == obs_json["feels_like"]
        assert obs.heat_index == obs_json["heat_index"]
        assert obs.lightning_strike_count == obs_json["lightning_strike_count"]
        assert (
            obs.lightning_strike_count_last_1hr
            == obs_json["lightning_strike_count_last_1hr"]
        )
        assert (
            obs.lightning_strike_count_last_3hr
            == obs_json["lightning_strike_count_last_3hr"]
        )
        assert (
            obs.lightning_strike_last_distance
            == obs_json["lightning_strike_last_distance"]
        )
        assert (
            obs.lightning_strike_last_epoch == obs_json["lightning_strike_last_epoch"]
        )
        assert obs.precip == obs_json["precip"]
        assert obs.precip_accum_last_1hr == obs_json["precip_accum_last_1hr"]
        assert obs.precip_accum_local_day == obs_json["precip_accum_local_day"]
        assert (
            obs.precip_accum_local_day_final == obs_json["precip_accum_local_day_final"]
        )
        assert (
            obs.precip_accum_local_yesterday == obs_json["precip_accum_local_yesterday"]
        )
        assert (
            obs.precip_accum_local_yesterday_final
            == obs_json["precip_accum_local_yesterday_final"]
        )
        assert (
            obs.precip_analysis_type_yesterday
            == obs_json["precip_analysis_type_yesterday"]
        )
        assert obs.precip_minutes_local_day == obs_json["precip_minutes_local_day"]
        assert (
            obs.precip_minutes_local_yesterday
            == obs_json["precip_minutes_local_yesterday"]
        )
        assert (
            obs.precip_minutes_local_yesterday_final
            == obs_json["precip_minutes_local_yesterday_final"]
        )
        assert obs.pressure_trend == obs_json["pressure_trend"]
        assert obs.relative_humidity == obs_json["relative_humidity"]
        assert obs.sea_level_pressure == obs_json["sea_level_pressure"]
        assert obs.solar_radiation == obs_json["solar_radiation"]
        assert obs.station_pressure == obs_json["station_pressure"]
        assert obs.timestamp == obs_json["timestamp"]
        assert obs.uv == obs_json["uv"]
        assert obs.wet_bulb_globe_temperature == obs_json["wet_bulb_globe_temperature"]
        assert obs.wet_bulb_temperature == obs_json["wet_bulb_temperature"]
        assert obs.wind_avg == obs_json["wind_avg"]
        assert obs.wind_chill == obs_json["wind_chill"]
        assert obs.wind_direction == obs_json["wind_direction"]
        assert obs.wind_gust == obs_json["wind_gust"]
        assert obs.wind_lull == obs_json["wind_lull"]


def test_convert_json_to_observation_no_precip(rest_station_observation_no_precip):
    obs_data = ObservationStationREST.from_dict(rest_station_observation_no_precip)

    assert isinstance(obs_data, ObservationStationREST)
    assert obs_data.elevation == rest_station_observation_no_precip["elevation"]
    assert obs_data.is_public == rest_station_observation_no_precip["is_public"]
    assert obs_data.latitude == rest_station_observation_no_precip["latitude"]
    assert obs_data.longitude == rest_station_observation_no_precip["longitude"]
    assert obs_data.public_name == rest_station_observation_no_precip["public_name"]
    assert obs_data.station_id == rest_station_observation_no_precip["station_id"]
    assert obs_data.station_name == rest_station_observation_no_precip["station_name"]
    assert obs_data.timezone == rest_station_observation_no_precip["timezone"]

    assert obs_data.obs[0].precip_accum_local_day_final == 0

    assert obs_data.obs[0].wind_cardinal_direction == "ENE"
    assert obs_data.obs[0].wet_bulb_globe_temperature_flag == WetBulbFlag.NONE
    assert obs_data.obs[0].wet_bulb_globe_temperature_category == 0
    assert obs_data.obs[0].uv_index_exposure == "high"
    assert obs_data.obs[0].uv_index_color == "orange"

    for obs, obs_json in zip(obs_data.obs, rest_station_observation_no_precip["obs"]):
        assert obs.air_density == obs_json["air_density"]
        assert obs.air_temperature == obs_json["air_temperature"]
        assert obs.barometric_pressure == obs_json["barometric_pressure"]
        assert obs.brightness == obs_json["brightness"]
        assert obs.delta_t == obs_json["delta_t"]
        assert obs.dew_point == obs_json["dew_point"]
        assert obs.feels_like == obs_json["feels_like"]
        assert obs.heat_index == obs_json["heat_index"]
        assert obs.lightning_strike_count == obs_json["lightning_strike_count"]
        assert (
            obs.lightning_strike_count_last_1hr
            == obs_json["lightning_strike_count_last_1hr"]
        )
        assert (
            obs.lightning_strike_count_last_3hr
            == obs_json["lightning_strike_count_last_3hr"]
        )
        assert (
            obs.lightning_strike_last_distance
            == obs_json["lightning_strike_last_distance"]
        )
        assert (
            obs.lightning_strike_last_epoch == obs_json["lightning_strike_last_epoch"]
        )
        assert obs.precip is None
        assert obs.precip_accum_last_1hr == obs_json["precip_accum_last_1hr"]
        assert obs.precip_accum_local_day == 0.0
        assert obs.precip_accum_local_day_final == 0
        assert (
            obs.precip_accum_local_yesterday == obs_json["precip_accum_local_yesterday"]
        )
        assert obs.precip_accum_local_yesterday_final == 0
        assert obs.precip_analysis_type_yesterday == 0
        assert obs.precip_minutes_local_day == obs_json["precip_minutes_local_day"]
        assert (
            obs.precip_minutes_local_yesterday
            == obs_json["precip_minutes_local_yesterday"]
        )
        assert obs.precip_minutes_local_yesterday_final == 0
        assert obs.pressure_trend == obs_json["pressure_trend"]
        assert obs.relative_humidity == obs_json["relative_humidity"]
        assert obs.sea_level_pressure == obs_json["sea_level_pressure"]
        assert obs.solar_radiation == obs_json["solar_radiation"]
        assert obs.station_pressure == obs_json["station_pressure"]
        assert obs.timestamp == obs_json["timestamp"]
        assert obs.uv == obs_json["uv"]
        assert obs.wet_bulb_globe_temperature == obs_json["wet_bulb_globe_temperature"]
        assert obs.wet_bulb_temperature == obs_json["wet_bulb_temperature"]
        assert obs.wind_avg == obs_json["wind_avg"]
        assert obs.wind_chill == obs_json["wind_chill"]
        assert obs.wind_direction == obs_json["wind_direction"]
        assert obs.wind_gust == obs_json["wind_gust"]
        assert obs.wind_lull == obs_json["wind_lull"]


def test_convert_json_to_observation_no_wind(rest_station_observation_no_wind):
    obs_data = ObservationStationREST.from_dict(rest_station_observation_no_wind)

    assert isinstance(obs_data, ObservationStationREST)

    obs = obs_data.obs[0]
    assert obs.wind_avg is None
    assert obs.wind_chill is None
    assert obs.wind_direction is None
    assert obs.wind_gust is None
    assert obs.wind_lull is None
    assert obs.feels_like is None

    # Properties should return None gracefully rather than raising
    assert obs.wind_cardinal_direction is None
    assert obs.wet_bulb_globe_temperature_flag == WetBulbFlag.NONE
    assert obs.wet_bulb_globe_temperature_category == 0


def test_convert_json_to_observation2(rest_station_observation2):
    obs_data = ObservationStationREST.from_dict(rest_station_observation2)
    assert obs_data.unknown_fields == {}

    assert isinstance(obs_data, ObservationStationREST)
    assert obs_data.elevation == rest_station_observation2["elevation"]
    assert obs_data.is_public == rest_station_observation2["is_public"]
    assert obs_data.latitude == rest_station_observation2["latitude"]
    assert obs_data.longitude == rest_station_observation2["longitude"]
    assert obs_data.public_name == rest_station_observation2["public_name"]
    assert obs_data.station_id == rest_station_observation2["station_id"]
    assert obs_data.station_name == rest_station_observation2["station_name"]
    assert obs_data.timezone == rest_station_observation2["timezone"]

    assert obs_data.obs[0].precip_accum_local_day_final == 0

    assert obs_data.obs[0].wind_cardinal_direction == "SSW"
    assert obs_data.obs[0].wet_bulb_globe_temperature_flag == WetBulbFlag.NONE
    assert obs_data.obs[0].wet_bulb_globe_temperature_category == 0
    assert obs_data.obs[0].uv_index_exposure == "low"
    assert obs_data.obs[0].uv_index_color == "green"

    for obs, obs_json in zip(obs_data.obs, rest_station_observation2["obs"]):
        assert obs.air_density == obs_json["air_density"]
        assert obs.air_temperature == obs_json["air_temperature"]
        assert obs.barometric_pressure == obs_json["barometric_pressure"]
        assert obs.brightness == obs_json["brightness"]
        assert obs.delta_t == obs_json["delta_t"]
        assert obs.dew_point == obs_json["dew_point"]
        assert obs.feels_like == obs_json["feels_like"]
        assert obs.heat_index == obs_json["heat_index"]
        assert obs.lightning_strike_count == obs_json["lightning_strike_count"]
        assert (
            obs.lightning_strike_count_last_1hr
            == obs_json["lightning_strike_count_last_1hr"]
        )
        assert (
            obs.lightning_strike_count_last_3hr
            == obs_json["lightning_strike_count_last_3hr"]
        )
        assert (
            obs.lightning_strike_last_distance
            == obs_json["lightning_strike_last_distance"]
        )
        assert (
            obs.lightning_strike_last_epoch == obs_json["lightning_strike_last_epoch"]
        )
        assert obs.precip == obs_json["precip"]
        assert obs.precip_accum_last_1hr == obs_json["precip_accum_last_1hr"]
        assert obs.precip_accum_local_day == obs_json["precip_accum_local_day"]
        assert obs.precip_accum_local_day_final == 0
        assert (
            obs.precip_accum_local_yesterday == obs_json["precip_accum_local_yesterday"]
        )
        assert obs.precip_accum_local_yesterday_final == 0
        assert (
            obs.precip_analysis_type_yesterday
            == obs_json["precip_analysis_type_yesterday"]
        )
        assert obs.precip_minutes_local_day == obs_json["precip_minutes_local_day"]
        assert (
            obs.precip_minutes_local_yesterday
            == obs_json["precip_minutes_local_yesterday"]
        )
        assert obs.precip_minutes_local_yesterday_final == 0
        assert obs.pressure_trend == obs_json["pressure_trend"]
        assert obs.relative_humidity == obs_json["relative_humidity"]
        assert obs.sea_level_pressure == obs_json["sea_level_pressure"]
        assert obs.solar_radiation == obs_json["solar_radiation"]
        assert obs.station_pressure == obs_json["station_pressure"]
        assert obs.timestamp == obs_json["timestamp"]
        assert obs.uv == obs_json["uv"]
        assert obs.wet_bulb_globe_temperature == obs_json["wet_bulb_globe_temperature"]
        assert obs.wet_bulb_temperature == obs_json["wet_bulb_temperature"]
        assert obs.wind_avg == obs_json["wind_avg"]
        assert obs.wind_chill == obs_json["wind_chill"]
        assert obs.wind_direction == obs_json["wind_direction"]
        assert obs.wind_gust == obs_json["wind_gust"]
        assert obs.wind_lull == obs_json["wind_lull"]


def test_convert_json_to_weather_data(rest_betterforecast_1):
    weather_data = WeatherDataForecastREST.from_dict(rest_betterforecast_1)

    # Assert that the conversion was successful
    assert isinstance(weather_data, WeatherDataForecastREST)

    assert isinstance(weather_data.current_conditions, CurrentConditions)
    assert isinstance(weather_data.forecast, Forecast)
    assert isinstance(weather_data.latitude, float)
    assert isinstance(weather_data.longitude, float)
    assert isinstance(weather_data.location_name, str)
    assert isinstance(weather_data.timezone, str)
    assert isinstance(weather_data.timezone_offset_minutes, int)
    assert isinstance(weather_data.units, ForecastUnits)


def test_convert_json_to_weather_data2(rest_betterforecast_2):
    try:
        weather_data = WeatherDataForecastREST.from_dict(rest_betterforecast_2)
    except Exception as e:
        pytest.fail(f"Failed to convert JSON data to WeatherData: {e}")

        # Assert that the conversion was successful
    assert isinstance(weather_data, WeatherDataForecastREST)

    assert isinstance(weather_data.current_conditions, CurrentConditions)
    assert isinstance(weather_data.forecast, Forecast)
    assert isinstance(weather_data.latitude, float)
    assert isinstance(weather_data.longitude, float)
    assert isinstance(weather_data.location_name, str)
    assert isinstance(weather_data.timezone, str)
    assert isinstance(weather_data.timezone_offset_minutes, int)
    assert isinstance(weather_data.units, ForecastUnits)


def test_convert_json_to_weather_data3(rest_betterforecast_3):
    try:
        weather_data = WeatherDataForecastREST.from_dict(rest_betterforecast_3)
    except Exception as e:
        pytest.fail(f"Failed to convert JSON data to WeatherData: {e}")

        # Assert that the conversion was successful
    assert isinstance(weather_data, WeatherDataForecastREST)

    assert isinstance(weather_data.current_conditions, CurrentConditions)
    assert isinstance(weather_data.forecast, Forecast)
    assert isinstance(weather_data.latitude, float)
    assert isinstance(weather_data.longitude, float)
    assert isinstance(weather_data.location_name, str)
    assert isinstance(weather_data.timezone, str)
    assert isinstance(weather_data.timezone_offset_minutes, int)
    assert isinstance(weather_data.units, ForecastUnits)


def test_convert_json_to_weather_data4(rest_betterforecast_4):
    try:
        weather_data = WeatherDataForecastREST.from_dict(rest_betterforecast_4)
    except Exception as e:
        pytest.fail(f"Failed to convert JSON data to WeatherData: {e}")

        # Assert that the conversion was successful
    assert isinstance(weather_data, WeatherDataForecastREST)

    assert isinstance(weather_data.current_conditions, CurrentConditions)
    assert isinstance(weather_data.forecast, Forecast)
    assert isinstance(weather_data.latitude, float)
    assert isinstance(weather_data.longitude, float)
    assert isinstance(weather_data.location_name, str)
    assert isinstance(weather_data.timezone, str)
    assert isinstance(weather_data.timezone_offset_minutes, int)
    assert isinstance(weather_data.units, ForecastUnits)


def test_convert_json_to_weather_data5(rest_betterforecast_5):
    try:
        weather_data = WeatherDataForecastREST.from_dict(rest_betterforecast_5)
    except Exception as e:
        pytest.fail(f"Failed to convert JSON data to WeatherData: {e}")

        # Assert that the conversion was successful
    assert isinstance(weather_data, WeatherDataForecastREST)

    assert isinstance(weather_data.current_conditions, CurrentConditions)
    assert isinstance(weather_data.forecast, Forecast)
    assert isinstance(weather_data.latitude, float)
    assert isinstance(weather_data.longitude, float)
    assert isinstance(weather_data.location_name, str)
    assert isinstance(weather_data.timezone, str)
    assert isinstance(weather_data.timezone_offset_minutes, int)
    assert isinstance(weather_data.units, ForecastUnits)
    assert weather_data.forecast.hourly[0].uv is None
    assert weather_data.forecast.hourly[0].feels_like is None


def test_convert_json_to_weather_data6(rest_betterforecast_6):
    try:
        weather_data = WeatherDataForecastREST.from_dict(rest_betterforecast_6)
    except Exception as e:
        pytest.fail(f"Failed to convert JSON data to WeatherData: {e}")

        # Assert that the conversion was successful
    assert isinstance(weather_data, WeatherDataForecastREST)

    assert isinstance(weather_data.current_conditions, CurrentConditions)
    assert isinstance(weather_data.forecast, Forecast)
    assert isinstance(weather_data.latitude, float)
    assert isinstance(weather_data.longitude, float)
    assert isinstance(weather_data.location_name, str)
    assert isinstance(weather_data.timezone, str)
    assert isinstance(weather_data.timezone_offset_minutes, int)
    assert isinstance(weather_data.units, ForecastUnits)
    assert weather_data.forecast.hourly[0].uv is None
    assert weather_data.forecast.hourly[0].feels_like is None


def test_hourly_forecast_missing_optional_fields(rest_betterforecast_missing_fields):
    """Test that hourly forecasts with missing sea_level_pressure or precip_probability parse without error."""
    try:
        weather_data = WeatherDataForecastREST.from_dict(
            rest_betterforecast_missing_fields
        )
    except Exception as e:
        pytest.fail(f"Failed to parse forecast with missing optional fields: {e}")

    hourly = weather_data.forecast.hourly
    assert len(hourly) == 3

    # First entry has both fields present
    assert hourly[0].sea_level_pressure == 1022.1
    assert hourly[0].precip_probability == 5

    # Second entry is missing both sea_level_pressure and precip_probability
    assert hourly[1].sea_level_pressure is None
    assert hourly[1].precip_probability is None

    # Third entry is missing only sea_level_pressure
    assert hourly[2].sea_level_pressure is None
    assert hourly[2].precip_probability == 0

    # ha_forecast should work even with None values
    forecasts = [h.ha_forecast for h in hourly]
    assert forecasts[1]["native_pressure"] is None
    assert forecasts[1]["precipitation_probability"] is None


def test_convert_weather_data_ha_forecast(rest_betterforecast_1):
    weather_data = WeatherDataForecastREST.from_dict(rest_betterforecast_1)
    forecasts_daily = [x.ha_forecast for x in weather_data.forecast.daily]
    forecasts_hourly = [x.ha_forecast for x in weather_data.forecast.hourly]

    assert forecasts_hourly[0]["datetime"] == "2023-12-28T18:00:00Z"
    assert forecasts_hourly[1]["datetime"] == "2023-12-28T19:00:00Z"
    assert forecasts_hourly[2]["datetime"] == "2023-12-28T20:00:00Z"
    assert forecasts_daily[0]["datetime"] == "2023-12-28T07:00:00Z"


def test_obs_st(obs_st_json):
    obs_json = DeviceObservationTempestREST.from_json(json.dumps(obs_st_json))
    obs_dict = DeviceObservationTempestREST.from_dict(obs_st_json)

    assert obs_dict == obs_json

    obs_st = DeviceObservationTempestREST.from_dict(obs_st_json)

    assert isinstance(obs_st, DeviceObservationTempestREST)

    assert obs_st.epoch == 1709057252
    assert obs_st.wind_lull == 0.77
    assert obs_st.wind_avg == 2.07
    assert obs_st.wind_gust == 3.98
    assert obs_st.wind_direction == 58
    assert obs_st.wind_sample_interval == obs_st_json["obs"][0][5]
    assert obs_st.pressure == obs_st_json["obs"][0][6]
    assert obs_st.air_temperature == obs_st_json["obs"][0][7]

    assert obs_st.status.status_code == obs_st_json["status"]["status_code"]
    assert obs_st.status.status_message == obs_st_json["status"]["status_message"]
    assert obs_st.device_id == obs_st_json["device_id"]
    assert obs_st.type == ObservationType.OBS_ST
    assert obs_st.source == obs_st_json["source"]
    assert obs_st.summary.pressure_trend == obs_st_json["summary"]["pressure_trend"]
    assert obs_st.summary.strike_count_1h == obs_st_json["summary"]["strike_count_1h"]
    assert obs_st.summary.strike_count_3h == obs_st_json["summary"]["strike_count_3h"]
    assert obs_st.summary.precip_total_1h == obs_st_json["summary"]["precip_total_1h"]
    assert obs_st.summary.strike_last_dist == obs_st_json["summary"]["strike_last_dist"]
    assert (
        obs_st.summary.strike_last_epoch == obs_st_json["summary"]["strike_last_epoch"]
    )
    assert (
        obs_st.summary.precip_accum_local_yesterday
        == obs_st_json["summary"]["precip_accum_local_yesterday"]
    )
    assert (
        obs_st.summary.precip_accum_local_yesterday_final
        == obs_st_json["summary"]["precip_accum_local_yesterday_final"]
    )
    assert (
        obs_st.summary.precip_analysis_type_yesterday == PrecipitationAnalysisType.NONE
    )
    assert obs_st.summary.feels_like == obs_st_json["summary"]["feels_like"]
    assert obs_st.summary.heat_index == obs_st_json["summary"]["heat_index"]
    assert obs_st.summary.wind_chill == obs_st_json["summary"]["wind_chill"]
    assert obs_st.epoch == obs_st_json["obs"][0][0]
    assert obs_st.wind_lull == obs_st_json["obs"][0][1]
    assert obs_st.wind_avg == obs_st_json["obs"][0][2]
    assert obs_st.wind_gust == obs_st_json["obs"][0][3]
    assert obs_st.wind_direction == obs_st_json["obs"][0][4]
    assert obs_st.wind_sample_interval == obs_st_json["obs"][0][5]
    assert obs_st.pressure == obs_st_json["obs"][0][6]
    assert obs_st.air_temperature == obs_st_json["obs"][0][7]
    assert obs_st.relative_humidity == obs_st_json["obs"][0][8]
    assert obs_st.illuminance == obs_st_json["obs"][0][9]
    assert obs_st.uv == obs_st_json["obs"][0][10]
    assert obs_st.solar_radiation == obs_st_json["obs"][0][11]
    assert obs_st.rain_accumulation == obs_st_json["obs"][0][12]
    assert obs_st.precipitation_type == PrecipitationType.NONE
    assert obs_st.average_strike_distance == obs_st_json["obs"][0][14]
    assert obs_st.strike_count == obs_st_json["obs"][0][15]
    assert obs_st.battery == obs_st_json["obs"][0][16]
    assert obs_st.report_interval == obs_st_json["obs"][0][17]
    assert obs_st.local_day_rain_accumulation == obs_st_json["obs"][0][18]
    assert obs_st.nc_rain_accumulation == obs_st_json["obs"][0][19]
    assert obs_st.local_day_nc_rain_accumulation == obs_st_json["obs"][0][20]
    assert obs_st.precipitation_analysis_type == PrecipitationAnalysisType.NONE


def test_rest_stations_endpoint(rest_stations_json, rest_stations_with_errors_json):
    station_data2 = StationsResponseREST.from_dict(rest_stations_with_errors_json)
    station_data1 = StationsResponseREST.from_dict(rest_stations_json)
    assert isinstance(station_data1, StationsResponseREST)
    assert isinstance(station_data2, StationsResponseREST)

    assert station_data1.station_device_map == {24432: [7654321, 123456]}
    assert station_data2.station_outdoor_device_map == {32364: [101382]}


def test_rest_station_endpoint(rest_station_json):
    station_data = StationsResponseREST.from_dict(rest_station_json)
    assert isinstance(station_data, StationsResponseREST)


def test_internal_error(rest_internal_error_json):
    status = Status.from_dict(rest_internal_error_json["status"])
    assert isinstance(status, Status)
